// -- File: BOND.CC

CLASS Bond
/* ----------------------------------------------------------------------
   Description:

   Encapsulates the standard securities-industry rules
           for the four calculations of interest to investors in
           bonds paying semiannual interest:
               findprice()  -- RETURNs the price given a yield
               findyield()  -- RETURNs the yield given a price
               findcoupon() -- RETURNs cost of first coupon
               finddate()   -- RETURNs date of first coupon
   Notes:
     See the accompanying "Bond.Txt" for a discussion of the nature of
     bonds and the bond market and for references for the formulas 
     used.	

     Percentages, such as the CPN property and the yield, are stated 
     as annual values * 100, so that a CPN of 6.25 means 6.25% 
     annually and a yield of 8.3204 means 8.3204% annualized. Price 
     and coupon cost are stated as dollars per $100 of face value of 
     the bond.

     The findyield() FUNCTION operates by trial and error calling the
     findroot() FUNCTION in the Root class found in Root.CC in the 
     the dUFLP library (and in ROOT.ZIP in the Compuserve Custom Class
     Library). Although that FUNCTION can yield other error RETURNs 
     when called in other contexts, the only one likely to be RETURNed
     here is -4, indicating that the initial bracket within which the 
     yield was assumed to fall was too narrow to include the actual 
     yield. If findyield() RETURNs a -4 result, set the range property
     of the bond object to a value larger than its default of .1 and 
     try again. A -4 result is likely only for bonds with a short term
     remaining to maturity or a price widely different from the RDV, 
     or both. If you evaluate such bonds often, you may wish to remove
     the range property and set the nError formula to something that 
     creates a larger bracket for such bonds, perhaps:

            nError = .1 * max(this.RDV/this.price, ;
                     this.price/this.RDV) ;
                     * max(1,540/this.DSM)				 				

   Programmers: 

     Philip Sydell 
     Jay Parsons 

   History: 

      10/30/1996 -- original
      11/25/1996 -- Minor modifications for dUFLP

   Instantiation: 

      set procedure to bond.cc additive
      _app.bond = new Bond()

    **** NOTE: If you do not use the release() method of Bond,
         you will not be able to release the reference to the
         Root class that is created in the init() method.

   Methods:
      More details on each of these is contained in the commentary for
      each method ...

   ** Note that this cc file uses Source Aliasing, specifically the
   ** dUFLP source Alias. If you do not know how to set up
   ** a Source Alias, see the instructions in WHATS.NEW at
   ** the top of the file, or in README.TXT.
   ** Any files referenced by :dUFLP:filename  must be
   ** included in an executable if you are building 
   ** one ... make sure that your project includes these
   ** files.
   **
   ** This particular program uses:
   **           root.cc
   **           duflp.h <== not referred to by source alias

   ---------------------------------------------------------------------- */

   #ifndef MB_OK
      #define MB_OK 0
   #endif

   PROTECT DSM, months, firstcoupondate, N, E, DSC, fracleft, ;
      firstcouponcost, sd, md, coupon, gooddata, calcdays, longprice

   * these three values must be given
   this.settledate = {//}       && date of settlement of the trade
   this.maturedate = {//}       && date of maturity of the bond
   this.CPN = 0			&& annual interest rate, as percent 

   * one of these is given, the other is found
   this.price = -10E9		&& price to pay per $100 of bonds		
   this.yield = -10E9           && yield to maturity as percent

   * these five may be overridden
   this.actual = .T.		&& actual/actual basis, else 30/360
   this.RDV = 100               && face or call value per $100
   this.accuracy = 0.0001       && accuracy of yield calculation
   this.maxiters = 100		&& maximum iterations for findroot()
   this.range = .1		&& size of starting bracket, by
                                && default a yield 90%-110% of
                                && the initial estimate.
  
   * these result from intermediate calculations and are protected
   this.DSM = 0			&& days settlement to maturity
   this.months = 0              && month portion of DSM
   this.firstcoupondate = {//}  && date first coupon is payable
   this.N = 0             	&& number of semiannual coupons
   this.E = 0 			&& days in coupon period in which
                                && settlement occurs 
   this.DSC = 0			&& days from settlement to first
                                && coupon date
   this.fracleft = 0 		&& fraction of E left, DSC/E 
   this.firstcouponcost = 0     && amount paid seller for first
                                && coupon, (1 - fracleft)*CPN/2
   this.iters = 0               && iterations so far, for testing
   this.sd = {//}		&& protected copy of this.settledate
   this.md = {//}               && protected copy of this.maturedate
   this.coupon = 0              && protected copy of this.CPN
   this.gooddata = .F. 		&& check on data validity
   
   CLASS::Init()
   
   FUNCTION Init
   /* -------------------------------------------------------------------
     Programmer..: Ken Mayer
     Date........: 12/09/1996
     Notes.......: Sets up root object
     Written for.: Visual dBASE 5.5
     Rev. History: 12/09/1996 -- Original Code
     Calls.......: root.cc
     Usage.......: Bond.Init()
     Example.....: _app.Bond = new Bond()  && init is called 
                                           && automatically
     Returns.....: Nul
     Parameters..: None
    ------------------------------------------------------------------- */
   
      PROTECT RootProc
      set procedure to :dUFLP:SetProc.prg additive
      if not SetProc( ":dUFLP:Root.CC" ) 
         set procedure to :dUFLP:root.cc additive
         this.RootProc = true
      else
         this.RootProc = false
      endif         
      this.Root        = new root()
      this.Root.Parent = this
      
   RETURN NULL

   FUNCTION Release
   /* -------------------------------------------------------------------
     Programmer..: Ken Mayer
     Date........: 12/09/1996
     Notes.......: Used to release this object from memory.
     Written for.: Visual dBASE 5.5
     Rev. History: 12/09/1996 -- Original Code
     Calls.......: none
     Usage.......: bond.Release()
     Example.....: _app.Bond.Release()
                   _app.Bond = NULL   && not required, but a good 
                                        && idea
     Returns.....: Null
     Parameters..: None
    ------------------------------------------------------------------- */

      this.Root.release()
      this.Root = null
      if this.RootProc
         close procedure :dUFLP:root.cc
      endif
      release object this

   RETURN NULL
   *-- EoM: Release()

   * find the cost of the first coupon (a protected property) 
   FUNCTION FindCoupon

      if this.calcdays()   
         RETURN this.firstcouponcost
      else
         RETURN -1
      endif
   * EoF: findcoupon
   
   * find date the first coupon matures (a protected property)
   FUNCTION finddate

      if this.calcdays()   
         RETURN this.firstcoupondate
      else
         RETURN {}
      endif
   * EoF: finddate
   
   * find the price given a yield
   FUNCTION findprice

      if .not. this.calcdays()
         RETURN -1
      endif

      if this.N = 1 
         this.price = 100 * (this.RDV + this.CPN/2) / ;
             (100 + this.DSM * this.yield/(2 * this.E)) ;
             - this.firstcouponcost
      else
         oB = this 
         this.price = this.longprice(this.yield)
      endif   

   RETURN this.price
   * EoF: findprice

   * find the yield given a price
   FUNCTION findyield

      if .not. this.calcdays()
         RETURN -1
      endif   

      if this.N = 1 

         * use the formula yield = 200*E/DSM*((RDV+CPN/2) / ;
                                   (PRICE + DCS*CPN/(2*E) -  1)      
         this.yield = 200*this.E/this.DSM * ((this.RDV + this.CPN/2) ;
                       / (this.price + this.firstcouponcost) - 1)
         RETURN this.yield  
         
      else

         * yield must be calculated by trial and error from the
         * price formula
         local guess, nerror, nRoot, nObj
         
         * approximate the yield by a formula 
         guess = 100 * ((this.CPN+2*(this.RDV-this.price)/this.N)) ;
                        / this.price 

         * and set starting bracket size for trials
         nError = guess*this.range
         
         * call findroot
         oB = this	&& sneaky "parameter" to longprice()
         
         this.Root.fp = class::longprice
         
         this.Root.x1 = guess + nError
         this.Root.x2 = guess - nError
         this.Root.offset = this.price
         this.yield = this.Root.findroot()
         if this.Root.flag # 1 .and. this.Root.flag # 2
            this.yield = - this.Root.flag
         endif
         this.iters = this.Root.iterations  && for testing
         RETURN this.yield

      endif
   * Eof: findyield

   * FUNCTION to find price of a bond with six months or more
   * to maturity, equivalent to SIA formula but faster. This
   * is called directly by findprice() and through findroot()
   * by findyield()
   FUNCTION longprice(yield)        && uses oB too

      local pVal, sTerm, sSum, k, y
      y = 1 + yield/200
      pVal = oB.RDV/y^(oB.N - 1 + oB.fracleft)
      sTerm = oB.CPN/2/y^oB.fracleft
      sSum = 0
      for k = 1 to oB.N
         sSum = sSum + sTerm
         sTerm = sTerm/y
      endfor

   RETURN pVal + sSum - oB.firstcouponcost    
   * Eof: longprice

   * FUNCTION to calculate days to maturity, number of coupons and
   * first coupon data
   FUNCTION calcdays

      * if calculation has been done already, just RETURN as before
      * if not done, save data
      if this.sd = this.settledate .and. this.md = this.maturedate ;
          .and. this.coupon = this.CPN
         RETURN this.gooddata
      else
         this.sd = this.settledate
         this.md = this.maturedate
         this.coupon = this.CPN
      endif
      * if properties are not set or wrongly set, RETURN with message
      if this.settledate = {//} .or. this.maturedate = {//} ;
          .or. this.CPN = 0
         msgbox("All of settledate, maturedate and CPN must be set",;
                "Bond::Calcdays Error",MB_OK)
         this.gooddata = .F.   
         RETURN this.gooddata
      endif
      if this.maturedate <= this.settledate 
         msgbox("Settlement must precede maturity",;
                "Bond::Calcdays Error",MB_OK)
         this.gooddata = .F.
         RETURN this.gooddata
      endif
 
      * first calculate full months to maturity
      this.months = 12*(year(this.maturedate) - year(this.settledate)) ;
                    + month(this.maturedate) - month(this.settledate) ;
                    - iif(day(this.maturedate) < ;
                    day(this.settledate),1,0)
      * then date first coupon is due; may be this.settledate
      this.firstcoupondate = this.addmonths(this.maturedate,;
                                            -6*int(this.months/6))

      * then number of coupons and first period data
      * if maturity is less than six months away, must use actual/
      * actual
      if this.actual .or. this.months < 6
         this.DSM = this.maturedate - this.settledate
         this.N = int(this.months/6) + 1
         this.E =  this.firstcoupondate - ;
                   this.addmonths(this.firstcoupondate,-6)
         this.DSC = this.firstcoupondate - this.settledate
      else
         this.DSM = this.days360(this.settledate, this.maturedate)
         this.N = int(this.DSM/180) + 1
         this.E = 180
         this.DSC = mod(this.DSM,180)
      endif   
      this.fracleft = this.DSC/this.E
      this.firstcouponcost = (1 - this.fracleft) * this.CPN/2   
      this.gooddata = .T.

   RETURN this.gooddata
   * EoF: calcdays
         
   * add a number of months to a date, or subtract. If the target
   * date does not exist, goes to first of next month.
   FUNCTION addmonths(dDate,nMonths)

      local dTry1,dTry2
      dTry1 = dDate - day(dDate) + 15 + 30.436875 * nMonths
      dTry2 = dTry1 - day(dTry1) + day(dDate)

   RETURN iif(month(dTry2) = month(dTry1),dTry2, ;
          dTry2 - day(dTry2) + 1)
   * EoF: addmonths

   * days between two dates using 30/360 basis
   FUNCTION days360(dDate1, dDate2)

      local nDays1, nDays2
      nDays1 = min(day(dDate1),30)
      nDays2 = day(dDate2) - iif(day(dDate2) = 31 .and. nDays1 =30,1,0)

   RETURN 360*(year(dDate2) - year(dDate1)) ;
          + 30*(month(dDate2) - month(dDate1));
          + nDays2 - nDays1
   * EoF: days360

ENDCLASS        
*-- End of Class:: BOND
         

/* -----------------------------------------------------------------------
   End of File: BOND.CC
   ----------------------------------------------------------------------- */