// -- File: FILEEX.CC

CLASS FileEx  // extended File class, except that this is not derived
              // from the built-in FILE class of VdBASE 7
/*
   ------------------------------------------------------------------
   Description:
 
       This class contains methods designed to assist working with
       files -- low-level file routines mostly, but not completely.
 
    Programmers: 
 
       Ken Mayer
 
    History: 
 
       11/25/1996 -- original version of these routines as a class.
       January 26, 1999 -- added FileInCurrentDir() 
       March 1, 1999 -- shortened IsDir() a bit and removed
                        IsNetDir() as unnecessary
       June 4, 2003 -- fixed some problems found in the delete()
                       method by Dan Barbiera.
                       
       January 8, 2018 -- Fixed a typo in the Locate method.  Mervyn Bick                  

    Instantiation: 
 
       set procedure to :dUFLP:fileex.cc additive
       oFile = new FileEx()
 
    Methods:
       More details on each of these is contained in the commentary 
       for each method ...
 
       Release()
         Releases this object from memory
 
       Delete()
         Deletes a given portion of a file
 
       GetLine()
         Extracts a line of text from a text file
  
       GetTempDir() 
         Assigns the DBTMP or TEMP directory to _app.TempDir

       InCurrentDir()
         Checks to see if a file exists in the current directory.
 
       Insert()
         Inserts specified number of NULLS into a low-level file
  
       IsDir()
         Checks for existance of a directory (local or network
         drive)
  
       Length()
         Returns length of file in bytes 
 
       Locate()
         Finds a string within a file
 
       ParsePath()
         Extracts and returns the path from a full path file spec.
 
       RemoveEof()
         Removes EOF marker from files
 
       Tell()
         Finds current position of file pointer 
 
       TempDir()
         Returns path of temporary directory as set from DOS
 
       TempFile()
         Returns a random filename
 
       TempName()
         Obtain a name for a temporary file of given extension that 
         does not conflict
 
       TextLine()
         Returns number of lines of text in an ASCII file
 
       TLine()
         Returns specific line of an ASCII text file
 
       TLineNo()
         Returns line number of phrase searching for in ASCII file
 
    Properties:
 
       None
   ------------------------------------------------------------------
*/

   FUNCTION Release
   /*
      -------------------------------------------------------------
       Programmer..: Ken Mayer
       Date........: 11/24/1996
       Notes.......: Used to release this object from memory.
       Written for.: Visual dBASE 5.5
       Rev. History: 11/24/1996 -- Original Code
       Calls.......: none
       Usage.......: fileex.Release()
       Example.....: oFile.Release()
                     oFile = NULL   // not required, but a good 
                                        // idea
       Returns.....: Null
       Parameters..: None
       ------------------------------------------------------------
   */
      release object this

   RETURN NULL
   // -- EoM: Release()

   FUNCTION Delete( nHandle, nBytes, nStart )
   /*
      -------------------------------------------------------------
       Programmer..: Adam L. Menkes
       Date........: 11/10/1996
       Notes.......: Deletes a given portion of a file. Taken from
                     TechNotes, April, 1992
                     Used to delete a portion of a file (text or 
                     binary) from the beginning of the file, the 
                     end of file or current pointer position. This 
                     routine accomplishes its task by writing the 
                     data you want to keep to a temp file, then 
                     overwriting the data you no longer want with 
                     the temp file. If you are on a network, make 
                     sure that you set TMP (or DBTMP) to either a 
                     local drive, or one where you have full 
                     rights.
       Written for.: dBASE IV, 1.5
       Rev. History: 04/01/1992 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this 
                                   class.
       Calls.......: fileex.TempFile()
       Usage.......: fileex.Delete(<nHandle>,<nBytes>,<nStart>)
       Example.....: nOpen = fopen("TEXT.TXT","RW")
                     ? oFile.Delete(nOpen,1000,1)
       Returns.....: Logical
       Parameters..: nHandle = file handle number, as returned by 
                               FOPEN
                     nBytes  = number of chars (bytes) to delete in 
                               file
                     nStart  = starting position, where:
                                0 is the beginning of the file
                                1 is the current file pointer 
                                  position
                                2 is the end of the file

      ***** NOTE: Attempting to use option 1 will return an error
            message, as this code cannot possibly function
            properly ...
      -------------------------------------------------------------
   */
      
      local nTemp,cTemp,nSave,nSeek,nRead,nWrite,lFlush,nClose
      
      // -- create a temporary file
      cTemp = class::TempFile("TMP")
      // -- save current position in file
      nSave = fseek(nHandle,0,1)
      
      do case
         case nStart = 0                  // beginning of file
              nSeek = fseek(nHandle, nBytes, 0)
              nTemp = fcreate(cTemp)
              do while not feof(nHandle)
                 nRead = fread(nHandle,254)
                 nWrite = fwrite(nTemp,nRead)
                 lFlush = fflush(nTemp)
              enddo
              nSeek = fseek(nTemp,0,0)
              nSeek = fseek(nHandle,0,0)
              do while not feof(nTemp)
                 nRead = fread(nTemp,254)
                 nWrite = fwrite(nHandle,nRead)
                 lFlush = fflush(nHandle)
              enddo
              nWrite = fwrite(nHandle,chr(0),0)
              nClose = fclose(nTemp)
              nSeek = fseek(nHandle,nSave,0)
              
         case nStart = 1                  // Current Location
/*              // -- skip these bytes
              nSeek = fseek(nHandle,nBytes,1)
              // -- write the rest to a temp file
              nTemp=fCreate(cTemp)
              do while not feof(nHandle)
                 nRead = fread(nHandle,32766)
                 nWrite = fwrite(nTemp,nRead)
                 lFlush = fflush(nTemp)
              enddo
              
              nSeek = fseek(nTemp,0,0)
              nSeek = fseek(nHandle,nSave,0)
              nWrite = fwrite(nHandle,chr(0),0)
              
              do while not feof(nTemp)
                 nRead = fread(nTemp,32766)
                 nWrite = fwrite(nHandle,nRead)
                 lFlush = fflush(nHandle)
              enddo
              nSeek = fseek(nHandle,nSave,0)
              nClose = fclose(nTemp)
*/
              msgbox( "Option 1 of the FileEx.delete() method does not "+;
                      "work properly. You may want to use FileEx32.cc instead.",;
                      "FileEx Error", 16 )
              erase (cTemp)
              return 1              
         case nStart = 2                  // End of File
              nSeek = fseek(nHandle,-1*abs(nBytes),2)
              nWrite = fwrite(nHandle,chr(0),0)
      endcase
      erase (cTemp)
      
   RETURN (ferror() = 0)
   // -- EoM: Delete()

   FUNCTION FileInCurrentDir( cFile )
   /*
       ------------------------------------------------------------
       Programmer..: Peter Rorlick
       Date........: January 25, 1999
       Notes.......: The file() function will check the current
                     directory as well as any in your path
                     statements, and if a file with the name
                     given is found, it will return a true, even
                     if the file is not in the current directory.
                     This was posted initially in the Visual
                     dBASE newsgroups by Peter, and stuffed into
                     the FileEX class by the librarian ...
       Written for.: Visual dBASE 7.01
       Rev. History: Jan. 25, 1999 -- first release
       Calls.......: None
       Usage.......: fileex.FileInCurrentDir( "somefile" )
       Example.....: oFile.FileInCurrentDir("mytable.dbf")
       Returns.....: logical
       Parameters..: cFile = name of file with extension
       ------------------------------------------------------------
   */
      local lReturn
      private cSavePath
      // save the current path settings:
      cSavePath = set("PATH")
      // turn off the path:
      set path to
      // now check to see if the file exists:
      lReturn = file( cFile )
      // re-set the path:
      set path to &cSavePath.
   return lReturn
   // -- EoM: FileInCurrentDir()

   FUNCTION GetLine( cFileName, cLookup, lCase, lEntire )
   /*
      -------------------------------------------------------------
       Programmer..: Adam L. Menkes
       Date........: 11/10/1996
       Notes.......: Used to extract a line of text from a text 
                     file.
       Written for.: dBASE IV, 1.5
       Rev. History: 04/01/1992 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: fileex.TLine()
                     fileex.TLineNo()
       Usage.......: fileex.GetLine(<cFileName>,<cLookup>,[<lCase>],;
                                     [<lEntire>])
       Example.....: ? oFile.GetLine("config.db","command",;
                                            false,false)
       Returns.....: A character expression
       Parameters..: cFileName = Name of file to extract text from
                     cLookup   = Text to look for
                     lCase     = Case sensitive? (Logical = true 
                                 or false)
                                 If empty, default is false
                     lEntire   = Return entire line, or the rest of
                                 line 
                                 true = return the entire line
                                 false = return everything following 
                                       cLookup
                                 If empty, default is true
      -------------------------------------------------------------
   */
      
      local nLine, cText
      
      // -- defaults
      lCase   = iif(pcount() <= 2,false,lCase)
      lEntire = iif(pcount() <=3,true,lEntire)
      // -- get the line ...
      nLine = CLASS::TLineNo(cFile,cLookup,lCase)
      cText = iif(nLine<=0,"",CLASS::TLine(cFile,nLine,lCase))
      cResult = upper(cText)
      
   RETURN iif( lEntire, cText, ;
          substr(cText,at(upper(cLookup),cResult)+len(cLookup)) )
   // -- EoM: GetLine()

   FUNCTION GetTempDir
   /*
      -------------------------------------------------------------
       Programmer..: Bowen Moursund
       Date........: 11/10/1996
       Notes.......: This routine checks the DOS environment for a 
                     variable pointing to a temporary (spill) 
                     directory, and assigns the return to 
                     _app.TempDir
       Written for.: dBASEWin, v5.0
       Rev. History: 09/12/1994
                     11/10/1996 -- Modified to work with this class.
       Calls.......: None
       Usage.......: fileex.GetTempDir()
       Example.....: oFile.GetTempDir()
       Returns.....: Path to Temp folder
       Parameters..: None
       Side Effects: Creates .TempDir property of _app
      -------------------------------------------------------------
   */
      
      local cTempDir
      cTempDir = GetEnv("DBTMP")
      cTempDir = iif("" <> cTempDir, cTempDir, GetEnv("TMP"))
      if "" <> cTempDir
         if right(cTempDir,1) = "\"
            cTempDir = stuff(cTempDir,rat("\",cTempDir),1,"")
        endif
      endif
      // assign to _app.
      _app.TempDir = cTempDir
      
   RETURN cTempDir
   // -- EoM: GetTempDir()


   FUNCTION Insert( nHandle, nBytes, nStart )
   /*
      -------------------------------------------------------------
       Programmer..: Adam L. Menkes
       Date........: 11/10/1996
       Notes.......: Inserts specified number of NULLS into a low-
                     level file. Taken from Technotes, April, 1992. 
                     Insert() works the way Delete() works, but 
                     in reverse.  See comments in Delete() about 
                     temp directory ...
       Written for.: dBASE IV, 1.5
       Rev. History: 04/01/1992 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: fileex.TempFile()
       Usage.......: fileex.Insert(<nHandle>,<nBytes>,<nStart>)
       Example.....: nOpen = fopen("TEST.TXT","RW")
                     ? oFile.Insert(nOpen,10,1)
       Returns.....: Logical
       Parameters..: nHandle = File Handle from FOPEN() function
                     nBytes  = Number of nulls to insert into file
                     nStart  = Location in file to start at, where:
                               0 = Beginning of file
                               1 = Current file pointer
                               2 = End of file
      -------------------------------------------------------------
   */
      local nTemp, cTemp, nSave, nRead, nWrite, nSeek, lFlush, nX, ;
            nClose
      
      cTemp = TempFile("ADM")         // create temp file name
      nSave = fseek(nHandle,0,1)      // save current position
      
      do case
         case nStart = 0           // beginning of file
              nTemp = fcreate(cTemp)
              nX = 1
              do while nX <= nBytes
                 nWrite = fwrite(nTemp,chr(0),1)
                 nX++
              enddo
              nSeek = fseek(nHandle,0,0)
              do while not feof(nHandle)
                 nRead = fread(nHandle,254)
                 nWrite = fwrite(nTemp,nRead)
                 lFlush = fflush(nTemp)
              enddo
              nSeek = fseek(nTemp,0,0)
              nSeek = fseek(nHandle,0,0)
              do while not feof(nTemp)
                 nRead = fread(nTemp,254)
                 nWrite = fwrite(nHandle,nRead)
                 lFlush = fflush(nHandle)
              enddo
              nWrite = fwrite(nHandle,chr(0),0)
              nclose = fclose(ntemp)
              nSeek = fseek(nHandle,0,0)
              
         case nStart = 1                  // current location
              // write the rest to a temp file
              nTemp = fcreate(cTemp)
              do while not feof(nHandle)
                 nRead = fread(nHandle,254)
                 nWrite = fwrite(nTemp,nRead)
                 lFlush = fflush(nTemp)
              enddo
              nSeek = fseek(nHandle,nSave,0)
              nX = 1
              do while nX <= nBytes
                 nWrite = fWrite(nHandle,chr(0),1)
                 nX++
              enddo
              nSeek = fseek(nTemp,0,0)
              do while not feof(nTemp)
                 nRead = fread(nTemp,254)
                 nWrite = fwrite(nHandle,nRead)
                 lFlush = fflush(nHandle)
              enddo
              nSeek = fseek(nHandle,nSave,0)
              nClose = fclose(nTemp)
               
         case nStart = 2                  // End of File
              nSeek = fseek(nHandle,0,2)
              nX = 1
              do while nX <= nBytes
                 nWrite = fwrite(nHandle,chr(0),1)
                 nX++
              enddo
      endcase
      erase (cTemp)
       
   RETURN (ferror() = 0)
   // -- EoM: Insert()

   FUNCTION IsDir( cDir )
   /*
      -------------------------------------------------------------
       Programmer..: Jim Sare 
                     Ken Mayer
                     Romain Strieff
       Date........: 02/24/1997
       Notes.......: Tests for existence of a directory
                     ** This version works on all OSes, including
                        Windows Vista and Windows 7 (write permission
                        issues).
       Written for.: dBASE IV v1.5+
       Rev. History: 12/20/1996
                     02/24/1997 - Shortened to much more simple
                                  form based on misc. discussions
                                  in the halls at Borland ...
                     03/01/1999 -- Shortened even more by 
                                  Romain -- this apparently
                                  works on network drives,
                                  and so obviates the need for
                                  IsNetDir().
       Calls.......: None
       Usage.......: fileex.IsDir( <cDir> )
       Example.....: if oFile.IsDir("C:\DBASE\DATA")
       Returns.....: logical true if directory is available, false 
                     if not
       Parameters..: cDir = [disk:] and directory to check for
      -------------------------------------------------------------
   */
   RETURN ( new array().dir( cDir+"\*.", "D" ) > 0 )
   // -- EoM: IsDir()

   FUNCTION Length( nHandle )
   /*
      -------------------------------------------------------------
       Programmer..: Rose Vines 
       Date........: 11/10/1996
       Notes.......: Finds length (in bytes) of a file and then 
                     returns the file pointer to its original 
                     position.
       Written for.: dBASE IV, 1.5
       Rev. History: 04/26/1993  -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: fileex.Tell()
       Usage.......: fileex.Length(<nHandle>)
       Example.....: ? oFile.Length( 6 )
       Returns.....: nLength   =  Length of file in bytes
       Parameters..: nHandle   =  dBASE file handle
      -------------------------------------------------------------
   */   
      local nCurPos, nLength

      // Locate current position in file without moving pointer
      nCurPos = CLASS::Tell( nHandle )

      // Find the length of the file by shifting the pointer to 
      // the end
      nLength = fseek( nHandle, 0, 2 )
      
      // Return the pointer to the original starting point
      fseek( nHandle, nCurPos, 0 )
      
   RETURN( nLength )
   // -- EoM: Length()

   FUNCTION Locate( nHandle, cSearch, lWantUpper )
   /*
      -------------------------------------------------------------
       Programmer..: Rose Vines (dBase version)
                     Matt Whelan (Clipper version - not included 
                      here)
       Date........: 11/10/1996
       Notes.......: Finds a string within a file, starting from 
                     current position of the file pointer (Operates
                     using low-level file functions).
       Written for.: dBASE IV, 1.5
       Rev. History: 04/26/1993  -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this 
                                   class.
       Calls.......: fileex.Tell()
                     fileex.Length()
       Usage.......: fileex.Locate(<nHandle>, <cSearch> ;
                                  [ ,<lWantUpper> ] )
       Example.....: oFile.Locate( 5, "CRABAPPLE COVE", true)
       Returns.....: nFoundPos  -  Starting position of the string 
                                   in file
       Parameters..: nHandle     =  dBASE file handle
                     cSearch     =  Search string
                     lWantUpper  =  Whether you want the text read
                                    converted to uppercase before
      		      comparison, defaults to false
      -------------------------------------------------------------
   */   
      local cBuffer, nCurPos, nStartPos, nBuffSize, nFlength
      local nBufPos, cTxtBuff, nBuffOffset, nFoundPos, cAddBuf
      
      nFoundPos  = -2
      
      // Convert buffer to uppercase if required
      if pcount() = 2
         lWantUpper = false
      endif

      // If a valid file handle has been passed...
      if nHandle > 0
         /*  Store our current position in the file,
             check the file length and then determine the
             buffer size. */
         nCurPos     = CLASS::Tell( nHandle )
         nStartPos   = nCurPos
         nFlength    = CLASS::Length( nHandle )
         nBuffSize   = min( 32766, nFlength )

         // Now start reading characters into the buffer
         do while nCurPos < nFlength
            cBuffer = ""
            do while len(cBuffer) < nBuffSize
               cAddBuf = fread( nHandle, 1 )
               // If you read in a null, replace it in the buffer
               // by a space
               if chr(0) = cAddBuf
                  cAddBuf = " "
               endif
               cBuffer = cBuffer + cAddBuf
            enddo

            if lWantUpper
               cBuffer = upper(cBuffer)
            endif

            // See if the search string is in the buffer
            nBufPos = at( cSearch, cBuffer )

            // and if it is, store its position in the file
            if nBufPos > 0
               nFoundPos = nCurPos + nBufPos - 1
               exit
            endif
         //   nCurPos = ftell( nHandle ) // Replaced. 2018/01/08 MB.
            nCurPos = class::tell( nHandle )
         enddo

         if nFoundPos < 1
            fseek( nHandle, nStartPos, 0 )
         else
            fseek( nHandle, nFoundPos, 0 )
         endif
      endif

   RETURN( nFoundPos )
   // EoM: Locate()

   FUNCTION ParsePath(cFullPath)
   /*
      -------------------------------------------------------------
       Programmer..: Bowen Moursund 
       Date........: 11/10/1996
       Notes.......: ParsPath() extracts and returns the path from 
                     a full path file specification.
       Written for.: dBASEWin v5.0
       Rev. History: 07/16/1992 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: None
       Usage.......: fileex.ParsPath( "<cFullPath>" )
       Example.....: set fullpath on
                     cDBF = dbf()
                     cPath = oFile.ParsPath( cDBF )
       Returns.....: The path only, including the trailing 
                     backslash, of the full path file specification
       Parameters..: cFullPath = a full path file spec,
                     e.g. "c:\dbase\dbase.exe"
      -------------------------------------------------------------
   */
     
      local cPath, nPos
      
      cPath = ""
      if "\" $ cFullPath
         nPos = 1
         do while left( right ( cFullPath, nPos ), 1 ) # "\"
            nPos++
         enddo
         cPath = substr(cFullPath, 1, len(cFullPath) - nPos + 1)
      endif
      
   RETURN cPath
   // -- EoM: ParsPath()

   FUNCTION RemoveEOF( cFile )
   /*
      -------------------------------------------------------------
       Programmer..: Kenneth Chan
       Date........: 11/10/1996
       Notes.......: Remove EOF marker from files
       Written for.: dBASE IV, 1.5
       Rev. History: 09/23/1993 1.0
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: None
       Usage.......: fileex.RemoveEOF( <cFile> )
       Example.....: oFile.RemoveEOF( "PRINTOUT.TXT" )
       Returns.....: NULL
       Parameters..: cFile      = name of file to truncate
      -------------------------------------------------------------
   */
      
      // Make sure file exists
      if file( cFile )
         hFile = fopen( cFile, "RW" )
         // If successful file open
         if hFile > 0
            // Goto last byte
            nPtr = fseek( hFile, -1, 2 )
            // If it's an EOF
            if fread( hFile, 1 ) = chr( 26 )
               // Reposition pointer
               nPtr = fseek( hFile, -1, 2 )
               // Truncate file
               nPtr = fwrite( hFile, "" )
            endif
            // Close file
            nPtr = fclose( hFile )
         endif
       endif
       
   RETURN NULL
   // -- EoM: RemoveEOF()

   FUNCTION Tell( nHandle )
   /*
      -------------------------------------------------------------
       Programmer..: Rose Vines
       Date........: 11/10/1996
       Notes.......: A shorthand way of finding the current 
                     position of file pointer in a file, without 
                     moving the pointer.
       Written for.: dBASE IV, 1.5
       Rev. History: 04/26/1993  -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: None
       Usage.......: fileex.Tell(<nHandle>)
       Example.....: oFile.Tell( 5 )
       Returns.....: Current position of pointer in a file.
       Parameters..: nHandle  =  dBASE file handle.
      -------------------------------------------------------------
   */

   RETURN( fseek( nHandle, 0, 1 ) )
   // -- EoM: Tell()

   FUNCTION TempDir
   /*
      -------------------------------------------------------------
       Programmer..: Adam L. Menkes
       Date........: 11/10/1996
       Notes.......: Returns path of temporary directory as set 
                     from DOS (i.e., SET DBTMP= ...) Taken from 
                     TechNotes, April,92
       Written for.: dBASE IV, 1.5
       Rev. History: 04/01/1992 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: None
       Usage.......: fileex.TempDir()
       Example.....: ? oFile.TempDir()
       Returns.....: Path of temporary directory
       Parameters..: None
      -------------------------------------------------------------
   */
       
      cTempDir = iif("" <> GetEnv("DBTMP"),GetEnv("DBTMP"),;
                           GetEnv("TMP"))
      
   RETURN cTempDir+iif(right(cTempDir,1)<> "\" and;
          left(os(),3) = "DOS" and not "" = cTempDir,"\","")
   // -- EoM: TempDir()

   FUNCTION TempFile(cFileExt)
   /*
      -------------------------------------------------------------
       Programmer..: Adam L. Menkes
       Date........: 11/10/1996
       Notes.......: Returns a random filename.
       Written for.: dBASEWin, v5.0
       Rev. History: 04/01/1992 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: fileex.TempDir()
       Usage.......: fileex.TempFile([cFileExt])
       Example.....: cVarFile = oFile.TempFile("$XY")
       Returns.....: Filename
       Parameters..: cFileExt = optional parameter - allows you to 
                                assign file extension to the end of 
                                the filename.
      -------------------------------------------------------------
   */
      
   RETURN CLASS::TempDir()+funique("TMP?????"+iif(pcount() = 0 or ;
           "" = cFileExt,"","."+ cFileExt) )
   // -- EoM: TempFile()

   FUNCTION TempName(cExt, lDBTMP)
   /*
      -------------------------------------------------------------
       Programmer..: Martin Leon 
       Date........: 11/10/1996
       Notes.......: Obtain a name for temporary file of a given 
                     extension that does not conflict with 
                     existing files.
       Written for.: dBaseWin, v5.0
       Rev. History: Originally part of Makestru(), 6-12-1991
                     04/26/1992 -- Jay Parsons, made a separate 
                                   function
                     05/27/1992 -- Bowen Moursund, added lDBTMP 
                                   option
                     02/22/1993 -- Minor update to PRIVATE command
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: fileex.GetTempDir()
       Usage.......: fileex.TempName( cExt , lDBTMP )
       Example.....: Sortfile = oFile.TempName( "DBF" , true )
       Returns.....: Name not already in use.
       Parameters..: cExt   = Extension of temporary ( without the 
                              ".")
                     lDBTMP = Optional. If true, function returns 
                              unique file name in the DBTMP 
                              subdirectory.
       Side Effects: The function will return a unique filename 
                     for the DEFAULT subdirectory if the lDBTMP 
                     option is used and the DOS environment 
                     variable DBTMP does not point to a valid 
                     subdirectory.
      -------------------------------------------------------------
   */
      
      local cFileName, cPath
      cPath = ""
      if lDBTMP
         if type("_app.TempDir") = "U"
            CLASS::GetTempDir()
          endif
          if "" <> _app.TempDir
             cPath = _app.TempDir + "\"
          endif
      endif
      do while true
         cFileName = funique( cPath + "TMP?????." + cExt )
         if upper( cExt ) # "DBF" or not ;
            ( file(cPath + left( cFileName, 9 ) + "MDX") or ;
              file (cPath + left( cFileName, 9 ) + "DBT") )
            exit
         endif
      enddo

   RETURN cFileName
   // -- EoM() TempName()

   FUNCTION TextLine( cTextFile )
   /*
      -------------------------------------------------------------
       Programmer..: Adam L. Menkes
       Date........: 11/10/1996
       Notes.......: Returns the number of lines of text in an 
                     ASCII Text file. Taken from TechNotes, April, 
                     1992
       Written for.: dBASE IV, 1.5
       Rev. History: 04/01/1992 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: None
       Usage.......: fileex.TextLine(<cTextFile>)
       Example.....: ? oFile.TextLine("CONFIG.DB")
       Returns.....: Number of lines
       Parameters..: cTextFile = name of file
      -------------------------------------------------------------
   */
      
      local nLines, nHandle, cTemp
      
      nLines = 0
      if file(cTextFile)   // if it exists ...
         nHandle = fopen(cTextFile,"R")
         do while not feof(nHandle)
            cTemp = fgets(nHandle,32766)
            nLines++
         enddo
         fclose(nHandle)
      endif
      
   RETURN nLines
   // -- EoM: TextLine()

   FUNCTION TLine( cTextFile, nLine )
   /*
      -------------------------------------------------------------
       Programmer..: Adam L. Menkes
       Date........: 11/10/1996
       Notes.......: Returns a specific line in an ASCII Text File. 
                     This is similar to the way MLINE() works on a 
                     memo field. Taken from TechNotes April, 1992.
       Written for.: dBASE IV, 1.5
       Rev. History: 04/01/1992 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: None
       Usage.......: fileex.TLine(<cTextFile>,<nLine>)
       Example.....: ? oFile.TLine("CONFIG.DB",20)
       Returns.....: Character expression - specified line of text 
                     file.
       Parameters..: cTextFile = name of text file
                     nLine     = line to return from text file
      -------------------------------------------------------------
   */
      
      local cText, nX, nHandle
      cText = ""
      nX = 1
      if file(cTextFile)    // if file exists ...
         nHandle = fopen(cTextFile,"R")
         do while not feof(nHandle)
            cText = fgets(nHandle,32766)
            if nX = nLine
               exit
            endif
            nX++
         enddo
         fclose(nHandle)
      endif
      
   RETURN cText
   // -- EoM: TLine()

   FUNCTION TLineNo( cTextFile, cLookup, lCase )
   /*
      -------------------------------------------------------------
       Programmer..: Adam L. Menkes
       Date........: 11/10/1996
       Notes.......: Returns the line number of the phrase you are
                     searching for in an ASCII Text File. This is 
                     similar to dBASE's AT() function, but works on 
                     LINES rather than CHARACTERS. Taken from 
                     TechNotes, April, 1992
       Written for.: dBASE IV, 1.5
       Rev. History: 04/01/1992 -- Original
                     Fall, 1994 -- revised for dBASE 5.0 for 
                                   Windows. Jay Parsons
                     11/10/1996 -- Modified to work with this class.
       Calls.......: None
       Usage.......: fileex.TLineNo(<cTextFile>,<cLookup>,[<lCase>])
       Example.....: ? oFile.TLineNo("CONFIG.DB","command",false)
       Returns.....: numeric value (the line number containing the 
                     line needed) returns -1 if not found
       Parameters..: cTextFile = Name of ASCII Text File
                     cLookup   = Text to search for ...
                     lCase     = Case Sensitive? (Default is false)
      -------------------------------------------------------------
   */
      
      local cPhrase, nHandle, cText, nX
      
      if pCount() = 3 and lCase
         lCase = true
         cPhrase = cLookup
      else
         lCase = false
         cPhrase = upper(cLookup)
      endif
      
      cText = ""
      nX = 1
      if file(cTextFile)
         nHandle = fopen(cTextFile,"R")
         do while not feof(nHandle)
            cText = fgets(nHandle,32766)
            if at(cPhrase,iif(lCase,cText,upper(cText))) > 0
               nClose = fclose(nHandle)
               RETURN nX
            endif
            nX++
         enddo
         
         fclose(nHandle)
      endif
      
   RETURN -1
   // -- EoM: TLineNo()

ENDCLASS
// -- End of Class: FileEx

/*
   ------------------------------------------------------------------
   End of File: FILEEX.CC
   ------------------------------------------------------------------
*/